from . import tf_bin_img, tf_bin_seq, tf_bin_tab, tf_bin_text, \
    tf_fore_img, tf_fore_tab, tf_multi_tab, tf_reg_tab, torch_bin_img, \
    torch_bin_tab, torch_multi_tab, torch_reg_tab

"""Register them all here so that they can be called like so: `from aiqc import tests`"""
